import itertools
import sys
import time
from typing import List, Optional

from pydantic import BaseModel, Extra, Field, PrivateAttr, validator


from chaiverse import config, errors, formatters, utils
from chaiverse.config import LEADERBOARD_ENDPOINT
from chaiverse.cli_utils.login_cli import auto_authenticate
from chaiverse.http_client import SubmitterClient
from chaiverse.lib.func import serialize_function
from chaiverse.schemas import DateRange


if 'ipykernel' in sys.modules:
    from IPython.core.display import display


class GenerationParams(BaseModel, extra=Extra.allow, title=''):
    temperature: float = Field(
        default=1.0,
        title="Temperature",
        description="Controls the randomness in the model's output. A higher temperature increases randomness."
    )
    top_p: float = Field(
        default=1.0,
        title="top_p",
        description="the model considers only the most probable tokens with probability above this threshold."
    )
    min_p: float = Field(
        default=0.0,
        title="min_p",
        description="Controls the minimum threshold a token must reach, relative to the top sampled token, in order for it to be sampled by your model."
    )
    top_k: int = Field(
        default=40,
        title="top_k",
        description="Number of highest probability tokens considered at each step of generation.")
    presence_penalty: Optional[float] = Field(
        default=0,
        title="presence_penalty",
        description="Encourages the model to use new and unique words by penalizing words that have already appeared in the text."
    )
    frequency_penalty: Optional[float] = Field(
        default=0,
        title="frequency_penalty",
        description="Encourages the model to use new and unique words by penalizing tokens based on their frequency in the text so far."
    )
    stopping_words: Optional[List[str]] = Field(
        default=["\n"],
        title="Stopping Words",
        description="Words or tokens that signal the model to stop generating further text.",
        # Determine how this field renders in HTML
        format="select2",
        unique_items=True,
        options={
            "select2": {
                "tokenSeparators": [" "],
            }
        }
    )
    max_input_tokens: Optional[int] = Field(
        default=config.DEFAULT_MAX_INPUT_TOKENS,
        title="Maximum input tokens",
        description="The maximum number of tokens the model that is fed into the model.")
    best_of: Optional[int] = Field(
        default=config.DEFAULT_BEST_OF,
        title="Best of",
        description="Number of completions produced by the model, which are used to prompt the reward model."
    )
    max_output_tokens: Optional[int] = Field(
        default=config.DEFAULT_MAX_OUTPUT_TOKENS,
        title="Maximum output tokens",
        description="The maximum number of tokens the model can generate in a single response.",
        options={
            "hidden": True
        }
    )

    _validation_errors: List[str] = PrivateAttr(default=[])

    def validate_currently_supported_format(self):
        # Validation is NOT done using pydantic validator for backward
        # compatibility
        self._validate_parameters()
        self._validate_presence_penalty()
        self._validate_frequency_penalty()
        self._validate_temperature()
        self._validate_top_k()
        self._validate_top_p()
        self._validate_min_p()
        self._validate_max_input_tokens()
        self._validate_max_output_tokens()
        self._validate_best_of()

        if self._validation_errors:
            error_messages = ', '.join(self._validation_errors)
            message = f"Generation Parameters validation failed with the following errors: {error_messages}"
            raise errors.ValidationError(message)

    def _validate_parameters(self):
        required_params = {'temperature', 'top_p', 'min_p', 'top_k', 'presence_penalty', 'frequency_penalty'}
        recognised_params = required_params.union({'max_input_tokens', 'stopping_words', 'best_of', 'max_output_tokens'})
        actual_params = set(self.dict(exclude_none=True).keys())

        # check there are no extra parameters
        extra_params = sorted(actual_params - recognised_params)
        if len(extra_params) > 0:
            msg = f'unrecognised generation parameters: {", ".join(extra_params)}'
            self._validation_errors.append(msg)

        # check there are no missing parameters
        missing_params = sorted(required_params - actual_params)
        if len(missing_params) > 0:
            msg = f'missing generation parameters: {", ".join(missing_params)}'
            self._validation_errors.append(msg)

    def _validate_presence_penalty(self):
        if self.presence_penalty is not None and not -2 <= self.presence_penalty <= 2:
            msg = f"`presence_penalty` must be in [-2, 2], got {self.presence_penalty}"
            self._validation_errors.append(msg)

    def _validate_frequency_penalty(self):
        if self.frequency_penalty is not None and not -2 <= self.frequency_penalty <= 2:
            msg = f"`frequency_penalty` must be in [-2, 2], got {self.frequency_penalty}"
            self._validation_errors.append(msg)

    def _validate_temperature(self):
        if self.temperature is not None and self.temperature < 0:
            self._validation_errors.append("`temperature` must be positive")

    def _validate_top_k(self):
        if self.top_k is not None and self.top_k <= 0:
            self._validation_errors.append("`top_k` must be greater than 0")

    def _validate_top_p(self):
        if self.top_p is not None and (self.top_p <= 0 or self.top_p > 1.0):
            self._validation_errors.append("`top_p` must be in (0, 1]")

    def _validate_min_p(self):
        if self.min_p is not None and (self.min_p < 0 or self.min_p > 1.0):
            self._validation_errors.append("`min_p` must be in [0, 1]")

    def _validate_best_of(self):
        if self.best_of is not None and (self.best_of <= 0 or self.best_of > 16):
            self._validation_errors.append("`best_of` must be in [1, 16]")

    def _validate_max_input_tokens(self):
        min_tokens = config.DEFAULT_MAX_INPUT_TOKENS
        max_tokens = 4096
        is_not_none = self.max_input_tokens is not None
        is_not_in_correct_range = not (min_tokens <= self.max_input_tokens <= max_tokens)
        if is_not_none and is_not_in_correct_range:
            msg = f"`max_input_tokens` must be in [{min_tokens}, {max_tokens}], got {self.max_input_tokens}"
            self._validation_errors.append(msg)

    def _validate_max_output_tokens(self):
        allowed_min = 32
        allowed_max = 112

        is_not_none = self.max_input_tokens is not None
        is_not_in_correct_range = not (allowed_min <= self.max_output_tokens <= allowed_max)

        if is_not_none and is_not_in_correct_range:
            msg = f"`max_output_tokens` must be in [{allowed_min}, {allowed_max}], got {self.max_output_tokens}"
            self._validation_errors.append(msg)


class FrontEndSubmissionRequest(BaseModel):
    class Config():
        title = 'Submit a model to Chaiverse'

    model_repo: str = Field(
        title="Model repository",
        description="HuggingFace repository hosting your model.",
        min_length=1
    )
    reward_repo: Optional[str] = Field(
        title="Reward model repository",
        description="HuggingFace repository hosting your reward model.",
        default=config.DEFAULT_REWARD_REPO
    )
    hf_auth_token: Optional[str] = Field(
        title="HuggingFace authentication token",
        description="A read-only HuggingFace token to access a private respository."
    )
    model_name: Optional[str] = Field(
        default=None,
        title="Model name",
        description="Set a custom name for your model."
    )
    generation_params: GenerationParams = Field(
        title="Generation Params",
        description="Set the parameters that control how your model samples tokens."
    )
    formatter: formatters.PromptFormatter = Field(
        title="Formatter",
        description="Provide a template to control how your model is prompted.",
        default=formatters.PygmalionFormatter()
    )
    reward_formatter: formatters.PromptFormatter = Field(
        title="Reward Formatter",
        description="Provide a template to control how your reward model is prompted.",
        default=formatters.PygmalionFormatter()
    )

    @validator("generation_params")
    def check_generation_params(cls, generation_params):
        generation_params.validate_currently_supported_format()
        return generation_params


class FrontEndBlendSubmissionRequest(BaseModel):
    submissions: List[str] = Field(..., description="submission_id for models to include in the blend.")
    model_name: Optional[str] = Field(None, description="name of your model.")


class ModelSubmitter:
    """
    Submits a model to the Guanaco service and exposes it to beta-testers on the Chai app.

    Attributes
    --------------
    developer_key : str
    verbose       : str - Print deployment logs

    Methods
    --------------
    submit(submission_params)
    Submits the model to the Guanaco service.

    Example usage:
    --------------
    submitter = ModelSubmitter(developer_key)
    submitter.submit(submission_params)
    """

    @auto_authenticate
    def __init__(self, developer_key=None, verbose=False):
        self.developer_key = developer_key
        self.verbose = verbose
        self._animation = self._spinner_animation_generator()
        self._progress = 0
        self._sleep_time = 0.5
        self._get_request_interval = int(10 / self._sleep_time)
        self._logs_cache = []

    def submit(self, submission_params):
        """
        Submits the model to the Guanaco service and wait for the deployment to finish.

        submission_params: dict
            model_repo: str - HuggingFace repo
            hf_auth_token (optional): str - A read-only token used to access model_repo
            generation_params: dict
                temperature: float
                top_p: float
                top_k: int
                repetition_penalty: float
            formatter (optional): PromptFormatter
            reward_formatter (optional): PromptFormatter
            model_name (optional): str - custom alias for your model
        """
        submission_params = submission_params.copy()
        submission_params = preprocess_submission(submission_params)
        return self._handle_submit(submit_model, submission_params)

    def submit_blend(self, submission_params):
        """
        Submits blend to the Guanaco service and wait for the deployment to finish.

        submission_params: dict
            submissions: list - submission_ids for the models in the blend
            model_name (optional): str - custom alias for your model
            override (optional): str - set parameter and formatter overrides for the sampled models
                {
                    "submission_id_v1": {
                        "generation_params": {
                            "temperature": 0.4
                        },
                        "formatter": {
                            "response_template": "{bot_name} (kind):"
                        }
                    }
                }
        """
        return self._handle_submit(submit_model_blend, submission_params)

    def submit_routed_blend(self, submission_params):
        return self._handle_submit(submit_model_routed_blend, submission_params)

    def submit_reward_blend(self, submission_params):
        return self._handle_submit(submit_model_reward_blend, submission_params)

    def _handle_submit(self, submission_func, submission_params):
        submission_response = submission_func(submission_params, self.developer_key)
        submission_id = submission_response.get('submission_id')
        self._print_submission_header(submission_id)
        status = self._wait_for_model_submission(submission_id)
        self._print_submission_result(status)
        self._progress = 0
        return submission_id

    def _wait_for_model_submission(self, submission_id):
        status = 'pending'
        while status not in {'deployed', 'failed', 'inactive'}:
            status = self._get_submission_status(submission_id)
            self._display_animation(status)
            time.sleep(self._sleep_time)
        return status

    def _get_submission_status(self, submission_id):
        self._progress += 1
        status = 'pending'
        if self._progress % self._get_request_interval == 0:
            model_info = get_model_info(submission_id, self.developer_key)
            self._print_latest_logs(model_info)
            status = model_info.get('status')
        return status

    def _spinner_animation_generator(self):
        animations = ['⣾', '⣽', '⣻', '⢿', '⡿', '⣟', '⣯', '⣷']
        return itertools.cycle(animations)

    def _display_animation(self, status):
        text = f" {next(self._animation)} {status}..."
        if 'ipykernel' in sys.modules:
            display(text, display_id="animation")
        else:
            print(text, end='\r')

    def _print_submission_header(self, submission_id):
        utils.print_color(f'\nModel Submission ID: {submission_id}', 'green')
        print("Your model is being deployed to Chaiverse, please wait for approximately 10 minutes...")

    def _print_submission_result(self, status):
        success = status == 'deployed'
        text_success = 'Model successfully deployed!'
        text_failed = 'Model deployment failed, please seek help on our Discord channel'
        text = text_success if success else text_failed
        color = 'green' if success else 'red'
        print('\n')
        utils.print_color(f'\n{text}', color)

    def _print_latest_logs(self, model_info):
        if self.verbose:
            logs = model_info.get("logs", [])
            num_new_logs = len(logs) - len(self._logs_cache)
            new_logs = logs[-num_new_logs:] if num_new_logs else []
            self._logs_cache += new_logs
            for log in new_logs:
                message = utils.parse_log_entry(log)
                print(message)


def preprocess_submission(submission_params):
    if submission_params.get('formatter'):
        submission_params["formatter"] = submission_params["formatter"].dict()
    if submission_params.get("reward_formatter"):
        submission_params["reward_formatter"] = submission_params["reward_formatter"].dict()
    return submission_params


@auto_authenticate
def submit_model(model_submission: dict, developer_key=None):
    http_client = SubmitterClient(developer_key)
    response = http_client.post(endpoint=config.SUBMISSION_ENDPOINT, json=model_submission)
    return response


@auto_authenticate
def submit_model_blend(model_submission: dict, developer_key=None):
    http_client = SubmitterClient(developer_key)
    response = http_client.post(endpoint=config.BLEND_SUBMISSION_ENDPOINT, json=model_submission)
    return response


@auto_authenticate
def submit_model_routed_blend(model_submission: dict, developer_key=None):
    model_submission = _serialize_tagging_function(model_submission)
    http_client = SubmitterClient(developer_key)
    response = http_client.post(endpoint=config.ROUTED_BLEND_SUBMISSION_ENDPOINT, json=model_submission)
    return response


def _serialize_tagging_function(model_submission):
    function = model_submission.get("tagging_function")
    if function:
        model_submission["tagging_function"] = serialize_function(function)
    return model_submission


@auto_authenticate
def submit_model_reward_blend(model_submission: dict, developer_key=None):
    http_client = SubmitterClient(developer_key)
    response = http_client.post(endpoint=config.REWARD_BLEND_SUBMISSION_ENDPOINT, json=model_submission)
    return response


@auto_authenticate
def get_model_info(submission_id, developer_key=None):
    http_client = SubmitterClient(developer_key)
    response = http_client.get(endpoint=config.INFO_ENDPOINT, submission_id=submission_id)
    return response


@auto_authenticate
def evaluate_model(submission_id, developer_key=None):
    http_client = SubmitterClient(developer_key)
    response = http_client.get(endpoint=config.EVALUATE_ENDPOINT, submission_id=submission_id)
    return response


@auto_authenticate
def get_my_submissions(developer_key=None):
    http_client = SubmitterClient(developer_key)
    response = http_client.get(endpoint=config.ALL_SUBMISSION_STATUS_ENDPOINT)
    return response


@auto_authenticate
def search_submissions(developer_key=None, **kwargs):
    http_client = SubmitterClient(developer_key)
    response = http_client.get(endpoint=config.SEARCH_SUBMISSIONS_ENDPOINT, params=kwargs)
    return response


@auto_authenticate
def deactivate_model(submission_id, developer_key=None):
    http_client = SubmitterClient(developer_key)
    response = http_client.get(endpoint=config.DEACTIVATE_ENDPOINT,submission_id=submission_id)
    print(response)
    return response


@auto_authenticate
def redeploy_model(submission_id, developer_key=None):
    http_client = SubmitterClient(developer_key)
    response = http_client.get(endpoint=config.REDEPLOY_ENDPOINT, submission_id=submission_id)
    print(response)
    return response


@auto_authenticate
def teardown_model(submission_id, developer_key=None):
    http_client = SubmitterClient(developer_key)
    response = http_client.get(endpoint=config.TEARDOWN_ENDPOINT, submission_id=submission_id)
    print(response)
    return response


def get_all_historical_submissions(developer_key):
    return get_submissions(developer_key=developer_key)


def get_submissions(developer_key=None, date_range:Optional[DateRange]=None):
    params = {}
    if date_range and date_range.start_date:
        params['start_date'] = f'{date_range.start_date}'
    if date_range and date_range.end_date:
        params['end_date'] = f'{date_range.end_date}'
    client = SubmitterClient(developer_key)
    resp = client.get(LEADERBOARD_ENDPOINT, params=params)
    return resp



import requests as requests_lib
from requests.adapters import HTTPAdapter, Retry

from chaiverse import config
from chaiverse.cli_utils.login_cli import get_developer_key_from_cache


# Increased retry from 3 to 6 only if initial connect failed, and it will not cause any side effects.
# This requests is intended to be directly imported to replace the requests library
requests = requests_lib.Session()
retries = Retry(connect=6, backoff_factor=1)
requests.mount('http://', HTTPAdapter(max_retries=retries))
requests.mount('https://', HTTPAdapter(max_retries=retries))


def get_url(endpoint, hostname=None, **kwarg):
    hostname = config.BASE_SUBMITTER_URL if hostname is None else hostname
    return (hostname + endpoint).format(**kwarg)


class _ChaiverseHTTPClient():
    def __init__(self, developer_key=None, hostname=None, debug_mode=False):
        if developer_key is None:
            developer_key = get_developer_key_from_cache()
        self.developer_key = developer_key
        self.hostname = hostname
        self.debug_mode = debug_mode

    @property
    def headers(self):
        headers = {"Authorization": f"Bearer {self.developer_key}"} if self.developer_key else {}
        return headers

    def get(self, endpoint, submission_id=None, **kwargs):
        url = get_url(endpoint, hostname=self.hostname, submission_id=submission_id)
        response = self._request(requests.get, url=url, **kwargs)
        return response

    def put(self, endpoint, **kwargs):
        url = get_url(endpoint, hostname=self.hostname)
        response = self._request(requests.put, url=url, **kwargs)
        return response

    def post(self, endpoint, submission_id=None, **kwargs):
        url = get_url(endpoint, hostname=self.hostname, submission_id=submission_id)
        response = self._request(requests.post, url=url, **kwargs)
        return response

    def delete(self, endpoint):
        url = get_url(endpoint, hostname=self.hostname)
        response = self._request(requests.delete, url=url)
        return response

    def _request(self, func, url, **kwargs):
        response = func(url=url, headers=self.headers, **kwargs)
        if not self.debug_mode:
            assert response.status_code == 200, response.json()
            response = response.json()
        return response


class SubmitterClient(_ChaiverseHTTPClient):
    def __init__(
            self,
            developer_key=None,
            hostname=None,
            debug_mode=False
    ):
        hostname = config.BASE_SUBMITTER_URL if not hostname else hostname
        super().__init__(developer_key, hostname, debug_mode)


class FeedbackClient(_ChaiverseHTTPClient):
    def __init__(
            self,
            developer_key=None,
            hostname=None,
            debug_mode=False
    ):
        hostname = config.BASE_FEEDBACK_URL if not hostname else hostname
        super().__init__(developer_key, hostname, debug_mode)


class PrometheusClient(_ChaiverseHTTPClient):
    def __init__(
            self,
            developer_key=None,
            hostname=None,
            debug_mode=False
    ):
        hostname = config.BASE_PROMETHEUS_URL if not hostname else hostname
        super().__init__(developer_key, hostname, debug_mode)


class AuthClient(_ChaiverseHTTPClient):
    def __init__(
            self,
            hostname=None,
            debug_mode=False
    ):
        hostname = config.BASE_AUTH_URL if not hostname else hostname
        super().__init__("", hostname, debug_mode)



from dataclasses import dataclass
from functools import lru_cache
from inspect import isclass
from typing import Optional, List
import sys

from pydantic import BaseModel, Field, validator


class _PromptFormatter(BaseModel):
    memory_template: str = Field(default = "{bot_name}'s Persona: {memory}\n####\n")
    prompt_template: str = Field(
        default="{prompt}\n<START>\n"
    )
    bot_template: str = Field(default="{bot_name}: {message}\n")
    user_template: str = Field(default="{user_name}: {message}\n")
    response_template: str = Field(default="{bot_name}:")
    truncate_by_message: Optional[bool] = Field(default=False)

    # TODO: Need to do this to test the XL reward model
    #       Consider actually deleting this, or reworking how the validation
    #       works so that it only happens on the submission

    #@validator("memory_template")
    #def validate_memory(cls, memory_template):
    #    if "{memory}" not in memory_template:
    #        raise ValueError("Formatter's memory_template must contain '{memory}'!")
    #    return memory_template

    #@validator("prompt_template")
    #def validate_formatter(cls, prompt_template):
    #    if "{prompt}" not in prompt_template:
    #        raise ValueError("Formatter's prompt_template must contain '{prompt}'!")
    #    return prompt_template

    #@validator("bot_template")
    #def validate_bot(cls, bot_template):
    #    if "{message}" not in bot_template:
    #        raise ValueError("Formatter's bot_template must contain '{message}'!")
    #    return bot_template

    #@validator("user_template")
    #def validate_user(cls, user_template):
    #    if "{message}" not in user_template:
    #        raise ValueError("Formatter's user_template must contain '{message}'!")
    #    return user_template


class PygmalionFormatter(_PromptFormatter):
    pass


class VicunaFormatter(_PromptFormatter):
    memory_template: str = "### Instruction:\n{memory}\n"
    prompt_template: str = "### Input:\n{prompt}\n"
    bot_template: str = "{bot_name}: {message}\n"
    user_template: str = "{user_name}: {message}\n"
    response_template: str = "### Response:\n{bot_name}:"


class ChatMLFormatter(_PromptFormatter):
    memory_template: str = "<|im_start|>system\n{memory}<|im_end|>\n"
    prompt_template: str = "<|im_start|>user\n{prompt}<|im_end|>\n"
    bot_template: str = "<|im_start|>assistant\n{bot_name}: {message}<|im_end|>\n"
    user_template: str = "<|im_start|>user\n{user_name}: {message}<|im_end|>\n"
    response_template: str = "<|im_start|>assistant\n{bot_name}:"


@lru_cache
def get_available_formatters():
    formatters = {}
    current_module = sys.modules[__name__]
    for key in dir(current_module):
        cls = getattr(current_module, key)
        if isclass(cls) and issubclass(cls, _PromptFormatter) and ("PromptFormatter" not in key):
            key = key.replace("Formatter", "")
            formatters[key] = cls()
    return formatters


class PromptFormatter(_PromptFormatter):
    memory_template: str = Field(
        title="Memory template",
        description="A template controlling how your model handles a bot's permanent memory. Must contain `{memory}`.",
        default = "{bot_name}'s Persona: {memory}\n####\n",
        enum=[formatter.memory_template for formatter in get_available_formatters().values()],
        format="select2",
        options={
            "enum_titles": list(get_available_formatters().keys()),
            "select2": {
                "tags": True,
            }
        }
    )
    prompt_template: str = Field(
        title="Prompt template",
        description="A template controlling how your model handles a bot temporary prompt. Must contain `{prompt}'.",
        default="{prompt}\n<START>\n",
        enum=[formatter.prompt_template for formatter in get_available_formatters().values()],
        format="select2",
        options={
            "enum_titles": list(get_available_formatters().keys()),
            "select2": {
                "tags": True,
            }
        }
    )
    bot_template: str = Field(
        title="Bot message template",
        description="A template controlling how your model handles a bot's messages. Must contain `{bot_name}' and `{message}'.",
        default="{bot_name}: {message}\n",
        enum=[formatter.bot_template for formatter in get_available_formatters().values()],
        format="select2",
        options={
            "enum_titles": list(get_available_formatters().keys()),
            "select2": {
                "tags": True,
            }
        }
    )
    user_template: str = Field(
        title="User message template",
        description="A template controlling how your model handles the user's messages. Must contain `{user_name}' and `{message}'.",
        default="{user_name}: {message}\n",
        enum=[formatter.user_template for formatter in get_available_formatters().values()],
        format="select2",
        options={
            "enum_titles": list(get_available_formatters().keys()),
            "select2": {
                "tags": True,
            }
        }
    )
    response_template: str = Field(
        title="Bot response template",
        description="A template controlling how your model is prompted for a bot response. Must contain `{bot_name}'.",
        default="{bot_name}:",
        enum=[formatter.response_template for formatter in get_available_formatters().values()],
        format="select2",
        options={
            "enum_titles": list(get_available_formatters().keys()),
            "select2": {
                "tags": True,
            }
        }
    )
    truncate_by_message: Optional[bool] = Field(
        title="Truncate by message",
        description="Truncate the conversation history in the context window on a message-by-message basis, rather than a character-by-character basis.",
        default=False,
        format="checkbox"
    )



from datetime import datetime
import json
from typing import List, Literal, Optional, Union, Dict

from pydantic import BaseModel, Extra, Field, root_validator

from chaiverse import config
from chaiverse.lib.now import utcnow
from chaiverse.lib.pydantic_tools import get_fields_in_schema, UnionParser
from chaiverse.lib.date_tools import convert_to_us_pacific_date


class BaseLeaderboardRow(BaseModel, extra=Extra.allow):
    developer_uid: str
    submission_id: str
    model_name: Optional[str]
    model_group: Optional[str] = ""
    status: str
    timestamp: datetime = Field(default_factory=utcnow)

    # preference metrics
    num_battles: int = 0
    num_wins: int = 0
    celo_rating: Optional[float] = None

    # alignment scores
    propriety_score: Optional[float] = 0
    propriety_total_count: Optional[float] = 0

    @root_validator()
    def add_model_group(cls, values):
        if not values["model_group"]:
            values["model_group"] = values.get("model_repo", "")[:24]
        return values

    @property
    def display_name(self):
        name = self.model_name if self.model_name else self.submission_id
        return name

    @property
    def win_ratio(self):
        return self.num_wins / self.num_battles if self.num_battles > 0 else None

    @property
    def us_pacific_date(self):
        us_pacific_date = convert_to_us_pacific_date(self.timestamp).date()
        return us_pacific_date

    @property
    def ineligible_reason(self) -> Optional[str]:
        return None

    def can_auto_deactivate(self):
        is_deployed = self.status == 'deployed'
        has_enough_battles = self.num_battles >= config.AUTO_DEACTIVATION_MIN_NUM_BATTLES
        has_enough_propriety = self.propriety_total_count >= config.AUTO_DEACTIVATION_MIN_NUM_PROPRIETY
        return is_deployed and has_enough_battles and has_enough_propriety

    def all_fields_dict(self):
        fields = get_fields_in_schema(self.__class__)
        fields_dict = {key: getattr(self, key) for key in fields}
        return fields_dict

    def base_fields_dict(self):
        fields = BaseLeaderboardRow.__fields__.keys()
        fields_dict = {key: getattr(self, key) for key in fields}
        return fields_dict

    # pydantic 2.x implements model_dump. Existing dict() in pydantic 1.x doesn't serialize recursively
    def model_dump(self):
        return json.loads(self.json())


class BasicLeaderboardRow(BaseLeaderboardRow):
    submission_type: Literal['basic'] = Field(default='basic')
    model_repo: str
    model_architecture: Optional[str]
    reward_repo: str = None
    model_num_parameters: float = None
    best_of: int = 1
    max_input_tokens: int
    max_output_tokens: Optional[int]

    @property
    def language_model(self):
        return self.model_repo

    @property
    def reward_model(self):
        return self.reward_repo

    @property
    def model_size(self):
        size_gb = round(self.model_num_parameters / 1e9) if self.model_num_parameters else None
        size_gb = f'{size_gb}B'
        return size_gb

    @property
    def ineligible_reason(self) -> Optional[str]:
        reason = None
        if self.status != 'deployed' and self.num_battles < config.LEADERBOARD_STABLE_ELO_REQUIRED_BATTLES:
            reason = f'num_battles<{config.LEADERBOARD_STABLE_ELO_REQUIRED_BATTLES}'
        if self.status != 'deployed' and self.propriety_total_count < config.LEADERBOARD_REQUIRED_PROPRIETY:
            reason = f'propriety_total_count < {config.LEADERBOARD_REQUIRED_PROPRIETY}'
        if self.max_output_tokens != config.DEFAULT_MAX_OUTPUT_TOKENS and self.max_output_tokens != None:
            reason = f'max_output_tokens!={config.DEFAULT_MAX_OUTPUT_TOKENS}'
        if self.status not in ['inactive', 'deployed', 'torndown']:
            reason = 'model is not deployable'
        return reason


class BlendLeaderboardRow(BaseLeaderboardRow):
    submission_type: Literal['blend'] = Field(default='blend')
    submissions: List[str]

    @property
    def language_model(self):
        return ','.join(self.submissions)

    @property
    def reward_model(self):
        return 'random'

    @property
    def model_size(self):
        return 'n/a'


class RewardBlendLeaderboardRow(BaseLeaderboardRow):
    submission_type: Literal['reward_blend'] = Field(default='reward_blend')
    reward_repo: str

    submissions: List[str]
    num_samples: int

    @property
    def language_model(self):
        pseudo_language_model = ','.join(self.submissions)
        return pseudo_language_model

    @property
    def reward_model(self):
        return self.reward_repo

    @property
    def model_size(self):
        return 'n/a'


class TaggedSubmissionID(BaseModel):
    submission_id: str
    tags: Optional[List[str]] = None


class RoutedBlendLeaderboardRow(BaseLeaderboardRow):
    submission_type: Literal['routed_blend'] = Field(default='routed_blend')
    router: str
    tagged_submissions: List[TaggedSubmissionID]

    @property
    def language_model(self):
        tagged_submissions = []
        for tagged_submission in self.tagged_submissions:
            tags = '|'.join(tagged_submission.tags)
            tagged_submissions.append(f'{tagged_submission.submission_id}:{tags}')
        pseudo_language_model = ','.join(tagged_submissions)
        return pseudo_language_model

    @property
    def reward_model(self):
        return self.router

    @property
    def model_size(self):
        return 'n/a'


class BinaryRoutedBlendLeaderboardRow(BaseLeaderboardRow):
    submission_type: Literal['binary_routed_blend'] = Field(default='binary_routed_blend')
    router: str
    routing_config: Dict

    @property
    def language_model(self):
        default_submission = self.routing_config['default_submission']
        secondary_submission = self.routing_config['secondary_submission']
        threshold = self.routing_config['threshold']
        return '{} (threshold: {}), {}'.format(default_submission, threshold, secondary_submission)

    @property
    def reward_model(self):
        return self.router

    @property
    def model_size(self):
        return 'n/a'


LeaderboardRow = Union[
    BasicLeaderboardRow,
    BlendLeaderboardRow,
    RewardBlendLeaderboardRow,
    RoutedBlendLeaderboardRow,
]



from typing import List, Literal, Optional

from chaiverse.http_client import SubmitterClient
from chaiverse.cli_utils.login_cli import auto_authenticate
from chaiverse.schemas import Competition
from chaiverse import config


@auto_authenticate
def get_competitions(status: Optional[Literal['submitting', 'evaluating']]=None, developer_key=None) -> List[Competition]:
    submitter_client = SubmitterClient(developer_key=developer_key)
    response = submitter_client.get(config.COMPETITIONS_ENDPOINT, params=dict(status=status))
    competitions = [Competition(**competition) for competition in response['competitions']]
    return competitions


def get_competition_by_id(competition_id: str, developer_key: str) -> Competition:
    submitter_client = SubmitterClient(developer_key=developer_key)
    url = config.COMPETITION_ENDPOINT.format(competition_id=competition_id)
    response = submitter_client.get(url)
    competition = Competition(**response)
    return competition


def create_competition(competition: Competition, developer_key: str=None):
    submitter_client = SubmitterClient(developer_key=developer_key)
    url = config.COMPETITIONS_ENDPOINT
    competition_id = submitter_client.post(url, json=competition.model_dump())
    return competition_id


def update_competition(competition_id: str, competition: Competition, developer_key: str=None):
    submitter_client = SubmitterClient(developer_key=developer_key)
    url = config.COMPETITION_ENDPOINT.format(competition_id=competition_id)
    competition = submitter_client.put(url, json=competition.model_dump())
    competition = Competition(**competition)
    return competition


def enroll_submission(submission_id, competition_id=None, developer_key=None):
    submission_ids = add_or_remove_enrolled_competition_ids(
        submission_ids_to_add=[submission_id],
        submission_ids_to_remove=[],
        competition_id=competition_id,
        developer_key=developer_key
    )
    return submission_ids


def withdraw_submission(submission_id, competition_id=None, developer_key=None):
    submission_ids = add_or_remove_enrolled_competition_ids(
        submission_ids_to_add=[],
        submission_ids_to_remove=[submission_id],
        competition_id=competition_id,
        developer_key=developer_key
    )
    return submission_ids


def add_or_remove_enrolled_competition_ids(submission_ids_to_add, submission_ids_to_remove, competition_id=None, developer_key=None):
    competition = _get_competition_for_enrollment(competition_id)
    submitter_client = SubmitterClient(developer_key=developer_key)
    for submission_id in submission_ids_to_remove:
        url = config.COMPETITION_ENROLLED_SUBMISSION_IDS_ENDPOINT.format(submission_id=submission_id, competition_id=competition.competition_id)
        submission_ids = submitter_client.delete(url)
        print(f'Withdrawn {submission_id} from {competition.display_name}.')
    for submission_id in submission_ids_to_add:
        url = config.COMPETITION_ENROLLED_SUBMISSION_IDS_ENDPOINT.format(submission_id=submission_id, competition_id=competition.competition_id)
        submission_ids = submitter_client.post(url)
        print(f'Enrolled {submission_id} into {competition.display_name}.')
    return submission_ids


def _get_competition_for_enrollment(competition_id=None, developer_key=None):
    if competition_id is None:
        competitions = get_competitions(status='submitting')
        assert competitions, 'No competition in submitting status'
        competition = competitions[0]
    else:
        competition = get_competition_by_id(competition_id, developer_key=developer_key)
    return competition



from dataclasses import dataclass
import json
import os
from pathlib import Path
from typing import List, Optional

from pydantic import BaseModel, Field

from chaiverse.config import CHAT_ENDPOINT
from chaiverse.cli_utils.login_cli import auto_authenticate
from chaiverse.utils import print_color
from chaiverse.http_client import SubmitterClient
from chaiverse.submit import get_all_historical_submissions


REPO_PATH = os.path.dirname(os.path.abspath(__file__))
RESOURCE_DIR = os.path.join(REPO_PATH, 'resources', 'bot_config')


def get_bot_names():
    files = os.listdir(RESOURCE_DIR)
    bot_names = [file.replace('.json', '') for file in files if file.endswith('.json')]
    return bot_names


class ChatMessage(BaseModel):
    sender: str
    message: str = Field(format="textarea")


class SubmissionChatbot():
    """
    Chat with a model that has been submitted and deployed to Chaiverse

    Attributes
    --------------
    submission_id : str
    developer_key : str

    Methods
    --------------
    chat(bot_config_file_name, show_model_input=False)
    Launch a chat session with the bot profile specified in bot_config_file_name

    show_available_bots()
    Displays the list of bots that are available for chat that can be passed into chat()

    Example usage:
    --------------
    chatbot = ChatWithSubmission(submission_id)
    chatbot.chat('vampire_queen')
    """
    @auto_authenticate
    def __init__(self, submission_id, developer_key=None):
        self.submission_id = submission_id
        self.developer_key = developer_key

    def chat(self, bot_config_file_name, show_model_input=False):
        """
        Launch a chat session with the bot profile specified in bot_config_file_name

        bot_config_file_name: str - name of the bot profile to chat with
        show_model_input: bool - whether to display the raw model input to your language model
        """
        bot_config = self._get_bot_config(bot_config_file_name)
        bot = Bot(self.submission_id, self.developer_key, bot_config)
        self._print_greetings_header(bot_config)
        user_input = input("You: ")
        while user_input != 'exit':
            response = bot.get_response(user_input)
            self._print_bot_response(bot_config, response, show_model_input)
            user_input = input("You: ")

    @staticmethod
    def show_available_bots():
        avaliable_bots = get_available_bots()
        print_color('Available Bots:', 'yellow')
        print(avaliable_bots)

    @staticmethod
    def _get_bot_config(config_file_name):
        bot_config_path = os.path.join(RESOURCE_DIR, f'{config_file_name}.json')
        return BotConfig.from_json(bot_config_path)

    def _print_greetings_header(self, bot_config):
        print_color(f'\nChatting with {bot_config.bot_label}, reply "exit" to quit.\n', 'yellow')
        print_color(f'{bot_config.bot_label}: {bot_config.first_message}', 'green')

    def _print_bot_response(self, bot_config, response, show_model_input):
        bot_response = response['model_output']
        if show_model_input:
            model_input = response['model_input']
            print_color('<Begin model input>', 'yellow')
            print_color(model_input, 'cyan')
            print_color('<End model input>', 'yellow')
        print_color(f"{bot_config.bot_label}: {bot_response}", 'green')


@dataclass
class BotConfig:
    memory: str
    prompt: str
    first_message: str
    bot_label: str

    @classmethod
    def from_json(cls, path):
        with open(path, 'r') as f:
            data = json.load(f)
        return cls(**data)


class Bot:

    def __init__(
            self,
            submission_id,
            developer_key,
            bot_config,
            endpoint = CHAT_ENDPOINT
    ):
        self.submission_id = submission_id
        self.developer_key = developer_key
        self.bot_config = bot_config
        self.endpoint = endpoint
        self._chat_history = self._init_chat_history()

    def get_response(self, user_input, generation_params=None, formatter=None):
        response = self._get_raw_response(user_input, generation_params, formatter)
        model_output = response['model_output']
        self._update_chat_history(model_output, self.bot_config.bot_label)
        return response

    def _get_raw_response(self, user_input, generation_params=None, formatter=None, debug_mode=False):
        self._update_chat_history(user_input, 'user')
        payload = {
            "memory": self.bot_config.memory,
            "prompt": self.bot_config.prompt,
            "chat_history": self._chat_history[:],
            "bot_name": self.bot_config.bot_label,
            "user_name": "You"
        }
        if generation_params:
            payload.update({"generation_params": generation_params.dict()})
        if formatter:
            payload.update({"formatter": formatter.dict()})
        http_client = SubmitterClient(self.developer_key, debug_mode=debug_mode)
        response = http_client.post(endpoint=self.endpoint, submission_id=self.submission_id, timeout=20, json=payload)
        return response

    def _update_chat_history(self, message, sender):
        message = {"sender": sender, "message": message}
        self._chat_history.append(message)

    def _init_chat_history(self):
        return [{"sender": self.bot_config.bot_label, "message": self.bot_config.first_message}]


def get_available_models(developer_key):
    models = get_all_historical_submissions(developer_key=developer_key)
    available_models = [k for k, v in models.items() if v['status'] == 'deployed']
    sorted_available_models = sorted(available_models)
    return sorted_available_models


def get_available_bots():
    bots = os.listdir(RESOURCE_DIR)
    avaliable_bots = '\n'.join([bot.replace('.json', '') for bot in bots if bot.endswith('.json')])
    return avaliable_bots


def get_bot_config(bot_name):
    config_file_name = f'{bot_name}.json'
    bot_config_path = Path(RESOURCE_DIR) / config_file_name
    return BotConfig.from_json(bot_config_path.absolute())


def get_bot_response(messages, submission_id, bot_config, generation_params=None, formatter=None, developer_key=None):
    chai_bot = Bot(submission_id, developer_key, bot_config)
    for message in messages:
        chai_bot._update_chat_history(message.message, message.sender)
    user_message = messages[-1].message
    response = chai_bot.get_response(user_message, generation_params, formatter)["model_output"]
    return response
